

#define MASKED_TWSR  (TWSR&0xf8)

#define BOOT_TWBR_VALUE \
  ((F_CPU/F_BOOT-16) <= 511 ?  (F_CPU/F_BOOT-16)/2 :(        \
  (F_CPU/F_BOOT-16)/4 <= 511 ? (F_CPU/F_BOOT-16)/4/2 :(      \
  (F_CPU/F_BOOT-16)/16 <= 511 ? (F_CPU/F_BOOT-16)/16/2 :(    \
  (F_CPU/F_BOOT-16)/64 <= 511 ? (F_CPU/F_BOOT-16)/64/2 :     \
    255 ))))

#define BOOT_TWPS_VALUE \
  ((F_CPU/F_BOOT-16) <= 511 ?  0 :(     \
  (F_CPU/F_BOOT-16)/4 <= 511 ? 1 :(     \
  (F_CPU/F_BOOT-16)/16 <= 511 ? 2 :(    \
  (F_CPU/F_BOOT-16)/64 <= 511 ? 3 :     \
    3 ))))

#define BOOTFUNC   __attribute__((section(".bootloader")))

void bc_init()                                     BOOTFUNC;
uint8_t bc_wait(uint8_t cr)                        BOOTFUNC;
uint8_t bc_write_byte(uint8_t V)                   BOOTFUNC;
uint8_t bc_slaw(uint8_t rw)                        BOOTFUNC;
uint8_t bc_write_page(uint8_t *ptr, uint16_t addr) BOOTFUNC;
uint8_t bc_read_page(uint8_t *ptr, uint16_t addr)  BOOTFUNC;
void bc_close()                                    BOOTFUNC;
void self_program_page(uint16_t addr, uint8_t *buf) BOOTFUNC;
uint8_t self_program_from_external_eeprom(uint16_t addr, uint8_t *buf) BOOTFUNC;
void boot_from_external_eeprom()                   BOOTFUNC;


uint16_t bc_step;

void bc_init() 
{
  TWBR= BOOT_TWBR_VALUE;
  TWSR= BOOT_TWPS_VALUE;
}

uint8_t bc_wait(uint8_t cr)
{
   uint16_t timeout;
   // a 10 ms wait 
   timeout =10000;
   TWCR= cr;
   while(!(TWCR & _BV(TWINT)))  
     { timeout--; if (!timeout) return 1; _delay_us(10); }
   return 0;
}

uint8_t bc_write_byte(uint8_t V)
{
  uint8_t S;
  TWDR= V;
  if (bc_wait(_BV(TWINT) | _BV(TWEN))) return 1;
  S= MASKED_TWSR;
  if (S==0x20 || S==0x30 || S==0x38 || S==0x48 || S==0x00) return 1;
  return 0;
}


uint8_t bc_slaw(uint8_t rw)
{
  return bc_wait(_BV(TWINT) | _BV(TWEN)| _BV(TWSTA)) ||  
         bc_write_byte(BC_ADDR+rw);
}


uint8_t bc_write_page(uint8_t *ptr, uint16_t addr)
{
  uint8_t R,i;
  R= 1;
  bc_step= 0;
  if (bc_slaw(0)) goto stop;
  bc_step= 1;
  if (bc_write_byte(addr>>8) || bc_write_byte(addr)) goto stop;
  bc_step= 3;
  for(i=0;i<64;i++) if (bc_write_byte(ptr[i])) goto stop;
  R= 0;
stop:
  TWCR= _BV(TWINT) | _BV(TWSTO) | _BV(TWEN);
  return R;
}
uint8_t bc_read_page(uint8_t *ptr, uint16_t addr)
{
  uint8_t R,i, K, S;
  R= 1;
  bc_step= 10;
  if (bc_slaw(0)) goto stop;
  bc_step= 11;
  if (bc_write_byte(addr>>8) || bc_write_byte(addr)) goto stop;
  bc_step= 12;
  if (bc_slaw(1)) goto stop;
  bc_step= 13;
  for(i=0;i<64;i++) 
  {
     if (i==63) K= bc_wait( _BV(TWINT) | _BV(TWEN) );
           else K= bc_wait( _BV(TWINT) | _BV(TWEN) | _BV(TWEA) );
  bc_step= 14;
     if (K) goto stop;
     S= MASKED_TWSR;
     ptr[i]= TWDR;
  bc_step= 15;
     if (S==0x38 || S==0x48 || S==0x00)
        goto stop;
  }
  R= 0;
stop:
  TWCR= _BV(TWINT) | _BV(TWSTO) | _BV(TWEN);
  return R;
}
  
void bc_close()
{
   TWCR &= ~_BV(TWEN);
}


void self_program_page(uint16_t addr, uint8_t *buf)
{
  uint16_t i;
  uint8_t sreg;

  sreg = SREG;
  cli();
    
  eeprom_busy_wait ();
  boot_page_erase(addr);
  boot_spm_busy_wait();

  for (i=0;i<128;i+=2,buf+=2)
    boot_page_fill(addr + i, buf[0] | (buf[1]<<8));

  boot_page_write(addr);
  boot_spm_busy_wait();
  boot_rww_enable();

  SREG= sreg;
}

 // the buffer is a 128 byte buffer which doesn't hold any meaningful
 // data initially. It's passed as an argument because I don't want
 // to allocate space for it.
uint8_t self_program_from_external_eeprom(uint16_t addr, uint8_t *buf)
{
  if (bc_read_page(buf,addr))        return 1;
  if (bc_read_page(buf+64, addr+64)) return 1;
  self_program_page(addr, buf);
  return 0;
}

 // The highest page in the external eeprom is used for storing
 // information about which pages are to be reprogrammed.
 // Atmega328p has 32K of flash memory, with 128 bytes per page.
 // So, we have 256 pages. The array in the last page holds
 // 32 bytes. Each page is identified by a bit in this array.
 // page0 is at array location 0, bit 0,
 // page1                      0      1
 // ..
 // so on.
 // The RAM contents are irrelevant after bootup, so I simply
 // use the first available RAM location for the buffers.
 // 
 // The bootloader must not be reprogrammed thru this interface
 // because I didn't make any provisions for it, it won't work.
 // I don't do any checks here for it, because I will strip the
 // bootloader out of the main program while programming.
void boot_from_external_eeprom()
{
  uint8_t sreg;
  uint16_t page;
  uint8_t  *data_buffer;
  uint8_t  *mask_buffer;
  cli();

  data_buffer= (uint8_t*) RAMSTART;
  mask_buffer= data_buffer+ 128;
  // reset= (void*) 0x0000;  

    // each BC page is 64 bytes
  if (bc_read_page(mask_buffer, 0x7f80)) goto fail;

  for(page=0;page<256;page++)
  {
    if ((mask_buffer[page/8]&_BV(page%8))==0) continue;
    if (self_program_from_external_eeprom(page*128, data_buffer)) break;
  }
  bc_close();

fail:
  wdt_enable(WDTO_120MS);
  while(1) ;
}


void boot_wdt_disable() 
     __attribute__((naked)) __attribute__((section(".init3")));
 
void boot_wdt_disable()
{
  MCUSR = 0; 
  wdt_disable();
}
