#include "binary.h"

/***************************************************************************
 The new LCD interface. It uses 4 bit communication and shifts out the 
 signals using a 74HC595.

 RS (LCD pin 4) is connected to 595 pin 1 =  Q1
  E (LCD pin 6)                     pin 2 =  Q2

 D4 (LCD pin 11)                    pin 3 =  Q3
 D5 (LCD pin 12)                    pin 4 =  Q4
 D6 (LCD pin 13)                    pin 5 =  Q5
 D7 (LCD pin 14)                    pin 6 =  Q6
 
 In order to output byte UUUU_VVVV with RS=Z, I output the following
 sequence:

 0_UUUU_0_Z_0
 0_UUUU_1_Z_0     // E=1, load the upper nibble
 0_VVVV_0_Z_0
 0_VVVV_1_Z_0     //               lower nibble

 Pin 15 on the LCD is connected to VCC. The corresponding pin on the
 board is pin 7 of the 595. However, these are not connected. I simply
 broke off pin 7 from the IC socket.
 
 ***************************************************************************/


#define _LCD_E     0x04
#define _LCD_RS    0x02

static void lcd_byte(uint8_t a)
{
  uint8_t i;
  for(i=0;i<8;i++)
  {
    if (a&0x80) { SET(LCD_DATA); } else { CLR(LCD_DATA); }
    SET(LCD_CLOCK);
    _delay_us(1);
    CLR(LCD_CLOCK);
    a<<= 1;
  }
  SET(LCD_RCLOCK);
  _delay_us(1);
  CLR(LCD_RCLOCK);
}

static void lcd_command(uint8_t rs, uint8_t data, uint8_t long_delay)
{
  if (rs) rs= _LCD_RS;
     else rs= 0;
  lcd_byte( ((data&0xf0)>>1) | rs );
  lcd_byte( ((data&0xf0)>>1) | rs | _LCD_E );
  lcd_byte( ((data&0xf0)>>1) | rs );

  lcd_byte( ((data&0x0f)<<3) | rs );
  lcd_byte( ((data&0x0f)<<3) | rs | _LCD_E );
  lcd_byte( ((data&0x0f)<<3) | rs );

  // The document specifies 37us for completion for all commands,
  // except for CLEAR_DISPLAY and RETURN_HOME. Those have a 1.52ms
  // completion time.
  if (long_delay) _delay_ms(2);
            else  _delay_us(50); 
}

void lcd_clear() { lcd_command(0, 1, 1); }
void lcd_home() { lcd_command(0, 2, 1); }
void lcd_char(uint8_t chr) { lcd_command(_LCD_RS, chr, 0); }

void lcd_goto(uint8_t row, uint8_t col)
{
  uint8_t addr= row*0x40 + col;
  lcd_command(0,addr|(1<<7), 0);
}


void lcd_init()
{
  OUTPUT(LCD_DATA); OUTPUT(LCD_CLOCK); OUTPUT(LCD_RCLOCK);
  _delay_ms(100);  // wait for a while after power up
  lcd_command(0, B_0010_1000, 0); // Function set: 4 bits, 2 lines
         // The datasheet specifies this twice? Interesting.
  lcd_command(0, B_0010_1000, 0); // Function set: 4 bits, 2 lines
  lcd_command(0, B_0000_1100, 0); // Display ON, no cursor
  lcd_clear();
  lcd_command(0, B_0000_0110, 0); // Entry Mode, direction= right
}

void lcd_pstr(const uint8_t *K)
{
  while(1)
  {
    uint8_t Z;
    Z= pgm_read_byte(K);
    if (!Z) return ;
    lcd_char(Z);
    K++;
  }
}
