#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include "coth.h"
#include "worker.h"
#include "sock.h"
#include "gfx.h"
#include "ihex.h"

  // these have a lot of margin. default timeout for the server
  // socket on my esp8266 is 180 seconds, which is 3 minutes.
  // JOB_TIME specifies the timeout for a response to a single
  // packet, not the whole job, so it's plenty.
#define IDLE_TIME (30*1000)
#define JOB_TIME  (5*1000)

/*****************************************************************
 The main thread:
  - manages the GUI
  - manages timers
  - does file I/O
  - sets up the connection

 The worker thread:
  - transfers data to/from the device according to the
    commands given by the user
  - closes the connection if there is an error
  - gives feedback to the user on the progress of work

 The worker thread will suspend its execution when it needs to
 read some data from the device. The main thread will do one
 of the following: 
  - let the worker thread continue with its execution when
    some data comes in.
  - tell the worker thread that the operation is cancelled
  - tell the worker thread that the operation has timed out

 In the latter two cases, the operation should be cancelled.
 In the last case, the socket should be closed and the user
 should make another connection to the device.
 *****************************************************************/

int worker_busy;
static int job_cancelled;
static int wo_func;
static void *wo_p1, *wo_p2;
static uint32_t wo_i1, wo_i2;

static uint32_t wo_idle_timer;
static uint32_t wo_job_timer;

static coth_t *comain, *coworker;

extern int device_socket;

static int  main_pipe;

static uint8_t param[1024];
static uint8_t send_buf[2048];
static uint8_t recv_buf[2048];
static uint8_t send_buf_len;
static uint32_t last_cmd_ident=1;

static int is_cancelled()
{
  if (job_cancelled) { job_cancelled= 0;return 1; }
  return 0;
}

static void signal_worker_completion(uint32_t i1,uint32_t i2,void *p1,void *p2)
{
  uint8_t c;
  main_i1= i1;
  main_i2= i2;
  main_p1= p1;
  main_p2= p2;
  write(main_pipe,&c,1);
}


void worker_run(int func, void *p1, void *p2, uint32_t i1, uint32_t i2)
{
  if (worker_busy)
    worker_cancel_job();
  wo_func= func;
  wo_p1= p1;
  wo_p2= p2;
  wo_i1= i1;
  wo_i2= i2;
  wo_idle_timer= 0;
  wo_job_timer= 0;
  coth_switch(comain, coworker);
}

static void update_firmware(char *fn, uint32_t boot_area_start);

static int command(int cmd, int plen)
{
  int i,R,L;
  uint32_t cmd_ident, id;
  send_buf_len= 0;

  b64_encode_begin(send_buf,&send_buf_len);
  for(i=0;i<32;i+=8) b64_encode(last_cmd_ident>>i);
  b64_encode(cmd);
  for(i=0;i<plen;i++) b64_encode(param[i]);
  b64_encode_end();

  send_buf[send_buf_len++]= '\r';
  send_buf[send_buf_len++]= '\n';

  cmd_ident= last_cmd_ident++;

  if (device_socket==-1 || is_cancelled()) return -1;
  if (sock_send(device_socket, send_buf, send_buf_len))
    { close_device(0); return -2; }

again:
  L= 0;
  while(1)
  {
    wo_job_timer= 0;
    coth_switch(coworker,comain);
    if (device_socket==-1 || is_cancelled())
      { gfx_logview_printf(Wlog, "command cancelled\n"); 
       return -2; }
    R= sock_recv(device_socket, recv_buf+L, sizeof(recv_buf)-L);
    if (R<0) { gfx_logview_printf(Wlog, "sock_recv returned %d\n",R);
               close_device(0); return -2; }
    for(i=0;i<R;i++) if (recv_buf[L+i]=='\r') break;
    L+= i;
    if (i!=R) break;
  }
  L= b64_decode(recv_buf, L);

  if (L<5) { gfx_logview_printf(Wlog, "Message too short\n"); return -2; }
  id= 0;
  for(i=0;i<4;i++) id= id | (recv_buf[i]<<(i*8));
  if (id<cmd_ident) goto again;
  if (id>cmd_ident) 
   { gfx_logview_printf(Wlog, "device is out of sync\n");
     return -2; }

  for(i=4;i<L;i++) param[i-4]= recv_buf[i];
  return L-4;
}

static void run_ident(int show)
{
  int R;
  R= command('I', 0);
  if (R<0) { gfx_logview_printf(Wlog, "Ident failed\n"); }
  else { param[R]= 0; if (show) 
          gfx_logview_printf(Wlog, "Ident: %s\n", param); }
}

static int param_encode_uint32(uint32_t V, int len)
{
  int i;
  for(i=0;i<4;i++) { param[len++]= V; V>>= 8; }
  return len;
}

static void set_baud_rate(uint32_t rate)
{
  int i,R;
  uint32_t V;
  i= param_encode_uint32(rate, 0);
  R= command('B', i);
  if (R<0 || param[0]!=0) gfx_logview_printf(Wlog, "set_baud_rate failed\n"); 
  else     gfx_logview_printf(Wlog,"set baud rate to %d",rate);
}

static void get_baud_rate()
{
  int R;
  R= command('b', 0);
  if (R<0) { gfx_logview_printf(Wlog, "get_baud_rate failed\n"); }
  else {
     uint32_t Z;
     Z= param[0] | (param[1]<<8) | (param[2]<<16) | (param[3]<<24);
     gfx_logview_printf(Wlog,"baud rate is: %d\n",Z);
     signal_worker_completion(Z, 0, NULL, NULL);
  }
}
        

static void change_ap_settings(char *ssid, char *pwd)
{
  int R;
  strcpy(param, ssid);
  R= command('S', strlen(ssid));
  if (R<0 || param[0]==0xff) 
    { gfx_logview_printf(Wlog, "Changing conf.ssid failed\n"); return ;}
  gfx_logview_printf(Wlog,"conf.ssid changed\n");

  strcpy(param, pwd);
  R= command('P', strlen(pwd));
  if (R<0 || param[0]==0xff) 
    { gfx_logview_printf(Wlog, "Changing conf.pwd failed\n"); return ;}
  gfx_logview_printf(Wlog,"conf.pwd changed\n");
}

#if 0
static void run_test(int addr)
{
  int i,j,R;
  uint8_t rdata[64];
  i= 0;
  param[i++]= addr;
  param[i++]= addr>>8;
  for(j=0;j<64;j++)
    param[i++]= rdata[j]= rand();
  gfx_logview_printf(Wlog, "Writing 64 random bytes " 
                           "starting at address %02x\n", addr);
  R= command('>', i);
  if (R<0 || param[0]==0xff)
   { gfx_logview_printf(Wlog,
      "Writing to external EEPROM failed (%s)\n", R<0? "R<0" : "0xff");
     return ;}
  for(j=0;j<i;j++) param[j]= 0;
  i= 0;
  param[i++]= addr;
  param[i++]= addr>>8;
  gfx_logview_printf(Wlog, "Reading back the written chunk\n");
  R= command('<', i);
  if (R<0 || param[0]==0xff)
   { gfx_logview_printf(Wlog, "Reading from external EEPROM failed (%s)\n"
                  , R<0? "R<0" : "0xff");
     return ;}
  for(j=0;j<64;j++)
   if (param[j+1]!=rdata[j])
   { gfx_logview_printf(Wlog, "Data read back doesn't match the sent data at"
                              "offset %d\n", j);
     return ; }
  gfx_logview_printf(Wlog, "All OK\n");
}
#endif

static int write_external_eeprom(uint16_t addr, uint8_t *buffer)
{
  int i,R;
  i= 0;
  param[i++]= addr;
  param[i++]= addr>>8;
  memcpy(param+i, buffer, 64);

  R= command('>', i+64);
  msleep(10); // 5 milliseconds is the maximum wait time between page writes
              // for 24xx256  
  if (R<0 || param[0]==0xff) return 1;
  return 0;
}

int verify_flash(uint16_t addr, uint8_t *data)
{
  int R;
  int i;
  i= 0;
  param[i++]= addr;
  param[i++]= addr>>8;
  R= command('?', i);

  if (R<0) return 1;

  for(i=0;i<64;i++)
   if (param[i]!=data[i])
   { gfx_logview_printf(Wlog, "Data read back doesn't match the sent data at"
                              "offset %d\n", i);
     return 1; }
  return 0;
}

static void run_test(int addr)
{
  int i,j,R;
  uint8_t rdata[128];

  for(i=0;i<128;i++) rdata[i]= rand();

  gfx_logview_printf(Wlog, "Writing 128 bytes to address %02x\n",addr);

  if (write_external_eeprom(addr, rdata))
     { gfx_logview_printf(Wlog, "First half failed\n"); return ;}

  if (write_external_eeprom(addr+64, rdata+64))
     { gfx_logview_printf(Wlog, "Second half failed\n"); return ;}

  gfx_logview_printf(Wlog,"'Booting' 1 page from external EEPROM\n");
  i= 0;
  param[i++]= addr;
  param[i++]= addr>>8;
  R= command('!', i);
  if (R<0 || param[0]==0xff)
     { gfx_logview_printf(Wlog, "Failed\n"); return ;}

  gfx_logview_printf(Wlog,"Verifying data in flash\n");
  i= 0;
  param[i++]= addr;
  param[i++]= addr>>8;

  if (verify_flash(addr, rdata))
     { gfx_logview_printf(Wlog, "Failed at first half\n"); return ;}

  if (verify_flash(addr+64, rdata+64))
     { gfx_logview_printf(Wlog, "Failed at second half\n"); return ;}


  gfx_logview_printf(Wlog, "All OK\n");
}

static void worker_main()
{
  while(1)
  {
    wo_idle_timer= 0;
    while (!wo_func) coth_switch(coworker, comain);
    worker_busy= 1;
    switch(wo_func)
    {
    case 'I': run_ident(wo_i1); break;
    case 'A': change_ap_settings(wo_p1, wo_p2); break;
    //case 'T': run_test(wo_i1); break;
    case 'T': update_firmware("../firmware/funky.hex", 0x7000);
              break;
    case 'b': get_baud_rate(); break;
    case 'B': set_baud_rate(wo_i1); break;
    }
    wo_func= 0;
    worker_busy= 0;
  }
}

int worker_init()
{
  int fds[2];
  if (pipe(fds)) return -1;
  main_pipe= fds[1];
  comain= coth_main();
  coworker= coth_new(worker_main, 20*1024);
  job_cancelled= 0;
  wo_func= 0;
  worker_busy= 0;
  return fds[0];
}

void worker_cancel_job()
{
  if (!worker_busy) return ;
  job_cancelled= 1;
  coth_switch(comain, coworker);
}

void open_device(char *addr)
{
  if (device_socket!=-1) return ;
  device_socket= sock_connect(addr, 1011);
  if (device_socket<0) return ;
  gfx_monitor_file(device_socket, 30, 1, 0, 0);
}


void close_device(int canc)
{
  if (device_socket==-1) return ;
  gfx_stop_monitor_file(device_socket, 30);
  sock_close(device_socket);
  device_socket= -1;
  if (canc) worker_cancel_job();
}


void worker_read_data()
{
  coth_switch(comain, coworker);
}

void worker_time_elapsed(unsigned int ms)
{
  if (worker_busy)
    { wo_job_timer += ms;
      if (wo_job_timer>=JOB_TIME) {  gfx_logview_printf(Wlog,"cancelling job, it takes too long.\n"); worker_cancel_job(); } }
  else
    { wo_idle_timer += ms;
      if (wo_idle_timer>=IDLE_TIME)
          worker_run('I', 0, 0, 1, 0); }
}


static void update_firmware(char *fn, uint32_t boot_area_start)
{
  char *e;
  uint8_t *flash_data;
  uint8_t pagemask[64];
  int i,j,R;
  if (boot_area_start<0x7000 || boot_area_start>0x7fff)
  {
    gfx_logview_printf(Wlog,"Boot area start should be between "
                            "0x7000 and 0x7fff\n");
  }
  e= read_hex_file(fn, 32*1024, 0xff, &flash_data);
  if (e)
  {
    gfx_logview_printf(Wlog,"error reading hex file %s\n%s\n",fn,e);
    return ;
  }
  for(i=boot_area_start;i<0x7fff;i++)
    flash_data[i]= 0xff;

  memset(pagemask, 0, 64);
  gfx_logview_printf(Wlog, "Writing %s to external EEPROM\n",fn);
  for(i=0;i<32768;i+=128)
  {
    for(j=0;j<128;j++)
      if (flash_data[i+j]!=0xff) break;
    if (j==128) continue; // empty page, skip it
    pagemask[ (i/128)/8] |= 1<<((i/128)%8);
    if (write_external_eeprom(i, flash_data+i) ||
        write_external_eeprom(i+64, flash_data+i+64))
    {
      gfx_logview_printf(Wlog, "External EEPROM write failed at page %d\n",
                     i/128);
      return ;
    }
    msleep(10); // for the EEPROM to get itself together
  }
  gfx_logview_printf(Wlog, "Writing page list to external EEPROM\n");
  if (write_external_eeprom(0x7f80, pagemask))
  {
      gfx_logview_printf(Wlog, "Page mask write to external EEPROM failed\n");
      return ;
  }
  
  R= command('F', 0);
  if (R<0 || param[0]==0xff)
  {
      gfx_logview_printf(Wlog, "Boot command failed\n");
      return ;
  } 
  gfx_logview_printf(Wlog, "Disconnecting and rebooting now.\n"
                    "Please reconnect when booting done\n");

  close_device(0);
}

