/***********************************************************************
 This file implements the interface to the usbterm on the host side.
 It uses libusb to communicate with the device.
 ***********************************************************************/
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <libusb-1.0/libusb.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include "iface.h"


// These probably should be in the usbterm.h file
#define VENDOR_ID 0x6666
#define PROD_ID   0x1291 // see ../firmware/usbtiny.h

static libusb_device_handle *devh;

void msleep(uint32_t milliseconds)
{
  usleep(milliseconds * 1000);
}


static int usb_write(int cmd, int arg1, int arg2, void *buf, int len)
{
  return libusb_control_transfer(devh,
            LIBUSB_ENDPOINT_OUT | LIBUSB_REQUEST_TYPE_VENDOR,
		cmd, arg1, arg2, buf, len, 1000);
}

static int usb_simple(int cmd, int arg1, int arg2)
{
  return libusb_control_transfer(devh, 
            LIBUSB_ENDPOINT_OUT | LIBUSB_REQUEST_TYPE_VENDOR,
              cmd, arg1, arg2, 0, 0, 500);
}

static int usb_read(int cmd, int arg1, int arg2, void *buf, int len)
{
  return libusb_control_transfer(devh,
            LIBUSB_ENDPOINT_IN | LIBUSB_REQUEST_TYPE_VENDOR,
		cmd, arg1, arg2, buf, len, 1000);
}

static void print_libusb_error(char *op, int code)
{
   fprintf(stderr,"USBTERM: %s: libusb error (%d) %s\n", op, code,
                     libusb_error_name(code)); 
}

static int usb_exit(int r)
{
   if (devh)
   {
     libusb_release_interface(devh, 0);
     libusb_attach_kernel_driver(devh,0);
     libusb_close(devh);
   }
   libusb_exit(NULL);
   return r;
}


static void exit_print_libusb_error(char *op,int code)
{
   print_libusb_error(op,code);
   exit(usb_exit(1));
}

#if 0
static void usb_exipif(char *op,int R)
{
  if (R<0) exit_print_libusb_error(op, R);
}
#endif

static void usb_init(int debuglevel)
{
   int r;
   const struct libusb_version* ver;

   r= libusb_init(NULL);
   if (r) exit_print_libusb_error("init-lib", r);
   libusb_set_debug(NULL,debuglevel); // MAX level.
   ver= libusb_get_version();
   printf("USBTERM: Initialized libusb-%d.%d.%d.%d%s\n",
         ver->major, ver->minor, ver->micro, ver->nano, ver->rc);
}

static void usb_open()
{
   devh= libusb_open_device_with_vid_pid(0, VENDOR_ID, PROD_ID);
   if (!devh)
     exit(fprintf(stderr,"USBTERM: can't open device %04x:%04x\n",
                          VENDOR_ID, PROD_ID));
  libusb_detach_kernel_driver(devh,0);
  libusb_claim_interface(devh, 0);
  printf("USBTERM: Opened device.\n");
  // libusb_reset_device(devh);
}

void Uinit(int debuglevel)
{
  usb_init(debuglevel);
  usb_open();
}

void Ufinish()
{
  usb_exit(0);
}

int Uwrite2(uint8_t cmd, uint16_t arg1, uint16_t arg2, uint8_t *buffer, int len)
{
  return usb_write(cmd, 0, 0, buffer, len);
}

int Uwrite1(uint8_t cmd, uint16_t arg, uint8_t *buffer, int len)
{
  return Uwrite2(cmd, arg, 0, buffer, len);
}

int Uwrite0(uint8_t cmd, uint8_t *buffer, int len)
{
  return Uwrite2(cmd, 0, 0, buffer, len);
}

int Uread2(uint8_t cmd, uint16_t arg1, uint16_t arg2, uint8_t *buffer, int len)
{
  return usb_read(cmd, arg1, arg2, buffer, len);
}

int Uread1(uint8_t cmd, uint16_t arg, uint8_t *buffer, int len)
{
  return Uread2(cmd, arg, 0, buffer, len);
}

int Uread0(uint8_t cmd, uint8_t *buffer, int len)
{
  return Uread2(cmd, 0, 0, buffer, len);
}

int Usimple2(uint8_t cmd, uint16_t arg1, uint16_t arg2)
{
  return usb_simple(cmd, arg1, arg2);
}

int Usimple1(uint8_t cmd, uint16_t arg)
{
  return Usimple2(cmd, arg, 0);
}

int Usimple0(uint8_t cmd)
{
  return Usimple2(cmd, 0, 0);
}

